using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

namespace TemperaturePlusLCD
{
    /// <summary>
    /// 
    /// Author:  Rick Winscot
    /// Website: http://www.quilix.com
    /// 
    /// This class interfaces with the TMP102 Digital Temperature Sensor 
    /// Breakout - via the I2C bus.
    /// 
    /// Product Info:
    /// http://www.sparkfun.com/products/9418
    /// 
    /// Wiring diagram:
    /// http://wiring.org.co/learning/libraries/tmp102sparkfun.html
    /// 
    /// Wire-up the breakout to the Netduino as follows:
    /// SDA -> Analog Pin 4
    /// SCL -> Analog Pin 5
    /// GND -> GND
    /// 3V  -> 3V (DO NOT CONNECT TO THE 5V PIN AS THIS WILL DAMAGE THE TMP102)
    /// 
    /// </summary>
    public class TMP102 : IDisposable
    {
        /// <summary>
        /// I had all kinds of sweet IDE esplosions until I found a juicy bit about the 
        /// Wire library for Arduino only needing the 7 most significant bits for the 
        /// address 0x91 >> 1 = 0x48. I decided to give this a go... and voila!
        /// </summary>
        private const int I2CAddress = 0x48;

        /// <summary>
        /// Faster... or slower - whatev.
        /// </summary>
        private const int I2CTimeout = 1000;

        /// <summary>
        /// Value returned if we cannot read a value from the sensor.
        /// </summary>
        /// <remarks>
        /// This value has been chose as it is out of range for this universe.
        /// </remarks>
        public const double ERROR_VALUE = -500;

        /// <summary>
        /// ELECTRICAL CHARACTERISTICS - p.g. 3
        /// Serial Bus Active, SCL Frequency = 400kHz
        /// 
        /// SERIAL INTERFACE - p.g. 10
        /// "The TMP102 supports the transmission protocol for both fast (1kHz to 400kHz) and high-speed (1kHz to 3.4MHz) modes."
        /// 
        /// http://www.sparkfun.com/datasheets/Sensors/Temperature/tmp102.pdf
        /// </summary>
        private const int I2CClockRateKHz = 400;

        /// <summary>
        /// TMP102 Digital Temperature Sensor instance.
        /// </summary>
        private static I2CDevice tmp102 = new I2CDevice(new I2CDevice.Configuration(I2CAddress, I2CClockRateKHz));

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public double GetTemperature()
        {
            byte[] buffer = new byte[2];
            I2CDevice.I2CTransaction[] transaction = new I2CDevice.I2CTransaction[] { I2CDevice.CreateReadTransaction(buffer) };

            int result = tmp102.Execute(transaction, 1000);
            double temperature;

            if (result == 2)
            {
                temperature = buffer[0];
                temperature += (buffer[1] >> 4) * 0.0625;
                return temperature;
            }

            return ERROR_VALUE;
        }

        /// <summary>
        /// Used to track when Dispose is called.
        /// </summary>
        private bool disposed = false;

        /// <summary>
        /// Implement IDisposable.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);

            // Call to GC.SupressFinalize will take this object
            // off the finalization queue and prevent multiple
            // executions.
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Initiate object disposal.
        /// </summary>
        /// <param name="disposing">Flag used to determine if the method is being called by the runtime (false) or programmatically (true).</param>
        private void Dispose(bool disposing)
        {
            // Spin-lock check for disposal
            if (!this.disposed)
            {
                if (disposing)
                {
                    // Dispose your managed resources here.
                    if (tmp102 != null)
                    {
                        tmp102.Dispose();
                    }
                }

                // Dispose your unmanaged resource here.
                // ...

                // Done - prevent accidental or intentional Dispose calls.
                disposed = true;
            }
        }
    }
}
